#=======================================================================
#	$Id: WebEdit.pm,v 1.6 2005/11/25 22:00:38 pythontech Exp $
#	CGI application to show / edit resources in triple database.
#	Copyright (C) 2005  Python Technology Limited
#
#	This program is free software; you can redistribute it and/or
#	modify it under the terms of the GNU General Public License
#	as published by the Free Software Foundation; either version 2
#	of the License, or (at your option) any later version.
#
#	This program is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  
#	02111-1307, USA.
#=======================================================================
package PythonTech::Triples::WebEdit;
use strict;
use PythonTech::Conv qw(hq uq);
use Digest::MD5 qw(md5_hex);

#-----------------------------------------------------------------------
#	Create a new triples-editor
#	  db		Triple store
#	  dbroot	Root of triple storage file system tree
#	  editing_users	List of user ids allowed to edit
#	  viewing_users	List of user ids allowed to view
#	  dbqmap	Qprefix map
#	  baseUrl	Base to make URL from relative URIs
#-----------------------------------------------------------------------
sub new {
    my $class = shift;
    my $self = {@_};
    if (! $self->{'db'}) {
	if (! $self->{'dbroot'}) {
	    die "$class: Neither db nor dbroot configured\n";
	}
    }
    bless $self, $class;
    return $self;
}

#-----------------------------------------------------------------------
#	Handle a web query
#-----------------------------------------------------------------------
sub webquery {
    my($self, $query) = @_;

    #--- Access control
    # If open_browsing anyone can view, without login
    # Otherwise need login to identify themselves; they may then have
    # permission to view and maybe edit, depending on permission
    # lists.
    my $canedit = 0;
    if (! $self->{'open_browsing'}) {
	my $user = $query->require_login 
	    or return;
	die "Viewing not authorized\n"
	    unless &user_in_list($user, $self->{'viewing_users'});
	$canedit = &user_in_list($user, $self->{'editing_users'});
    }

    #--- Multi-database access
    # Currently hardcoded to DBHash
    if ($self->{'dbroot'}) {
	my $db = $query->param('db'); # relative path of database file
	die "Triple db query parameter not defined\n"
	  unless defined $db;
	die "Detected illegal .. in relative path [$db]\n"
	  unless (index($db,"../") < 0); 
	my $dbpath = $self->{'dbroot'} . $db;
	eval "require PythonTech::Triples::DBHash";
	die $@ if $@;
	my $hashdb = PythonTech::Triples::DBHash->new({
	    filename => "$dbpath",
	    access => 'w',
	});
	$self->{'db'} = $hashdb;
	$self->{'dbparam'} = $db; # keep db param for new queries
    }
    my $uri = $query->param('u');
    my $backuri = $query->param('back');
    $self->{'backparam'} = $backuri; # keep back param for new queries

    my $method = $query->request_method;
    if ($method eq 'GET') {
	if (defined $query->param('search')) {
	    $self->search($query);
	} else {
	    $self->show($query, $uri,$canedit);
	}
    } elsif ($method eq 'POST') {
	if (! $canedit) {
	    die "User not allowed to edit\n";
	} elsif ($query->param('delete')) {
	    $self->delete($query, $uri);
	    $self->show($query, $uri,$canedit);
	} elsif ($query->param('add')) {
	    $self->add($query, $uri,0);
	    $self->show($query, $uri,$canedit);
	} elsif ($query->param('add.lit')) {
	    $self->add($query, $uri,1);
	    $self->show($query, $uri,$canedit);
	} else {
	    die "Unknown POST request\n";
	}
    } else {
	die "Unsupported HTTP request method $method\n";
    }
}

#-----------------------------------------------------------------------
#       Sort compare, which gives "rdf:type" the lowest sort value
#-----------------------------------------------------------------------
sub typesfirst {
    ($a ne 'rdf:type') <=> ($b ne 'rdf:type') || $a cmp $b
}

#-----------------------------------------------------------------------
#	Show details of a URI
#-----------------------------------------------------------------------
sub show {
    my($self, $query,$uri,$editable) = @_;
    my $resp = $query->response;
    my $db = $self->{'db'};
    my $selfurl = $query->url(-path_info=>1);

    $resp->write($query->start_html(-title=>$uri));

    my $qmapper = $self->{'dbqmap'};
    my $url;			# URI may correspond to a URL
    if ($qmapper) {
	$url = $qmapper->decode($uri);
	$uri = $qmapper->encode($uri);
	my $last = $qmapper->lastq;
	if (defined $last) {
	    $resp->write("<h4> ".hq($last)." = ".hq($qmapper->decode($last))."</h4>\n");
	}
    } elsif ($uri =~ /^http:/) {
	$url = $uri;
    } elsif ($uri !~ /^[-\w]+:/ && $self->{'baseUrl'}) {
	$url = $self->{'baseUrl'} . $uri;
    }

    if ($editable) {
	$resp->write("<form method='POST'>
<input type='hidden' name='u' value=\"".hq($uri)."\" />\n");
	$resp->write($self->params_hidden);
    }
    my $backurl;
    if ($self->{'backparam'}) {
	$backurl = $self->{'backparam'};
    }

    # Only anchor if absolute http: URI
    my $hsubj;
    $hsubj = ($url =~ /^http:/) ?
      "<a href=\"" . hq($url) . "\">" . hq($uri) . "</a>" : 
	hq($uri);
    $resp->write(" <table border='1'>\n");
    if ($backurl) {
        my $hback = "<a href=\"".hq($backurl)."\">".hq($backurl)."</a>";
        $resp->write(" <tr><th colspan='2'>Model context: $hback</th></tr>\n");
    }
    $resp->write(" <tr><th colspan='2'>$hsubj</th></tr>\n");
    # Forward relations
    # N.B. s2p may be slow for large triplestores without index_s2p
    # 'iterate' should do the best and avoid multiple queries.
    my %by_pred;
    $db->iterate($uri,undef,undef, sub {
	my($subj,$pred,$obj) = @_;
	$by_pred{$pred}->{$obj} = 1;
	undef;			# Continue
    });
    my @predicates = sort typesfirst keys %by_pred;
    foreach my $p (@predicates) {
	my @o = sort keys %{$by_pred{$p}};
	my $c1 = "<td rowspan='".scalar(@o)."'>".hq($p)."</td>";
	foreach my $o (@o) {
	    my($sel,$val);
	    if (my($str) = $o =~ /^=(.*)/s) {
		$sel = "=" . md5_hex($str);
		$val = "<span class='literal'>".hq($str)."</span>";
	    } else {
		$sel = $o;
		$val = "<a href=\"".hq("$selfurl?u=".uq($o).$self->params_uq)."\">".hq($o)."</a>";
	    }
	    my $checkbox;
	    if ($editable) {
		$checkbox = "<input type='checkbox' name='po' value=\"".hq("$p $sel")."\" /> ";
	    }
	    $resp->write("<tr>$c1<td>$checkbox$val</td></tr>\n");
	    $c1 = "";
	}
    }
    # Reverse relations
    %by_pred = ();
    $db->iterate(undef,undef,$uri, sub {
	my($subj,$pred,$obj) = @_;
	$by_pred{$pred}->{$subj} = 1;
	undef;
    });
    foreach my $p (sort keys %by_pred) {
	my @s = sort keys %{$by_pred{$p}};
	my $c1 = "<td rowspan='".scalar(@s)."'>~ ".hq($p)."</td>";
	foreach my $s (@s) {
	    my $val;
	    if (my($str) = $s =~ /^=(.*)/s) {
		# Should never happen
		$val = "<span class='literal'>".hq($str)."</span>";
	    } else {
		$val .= "<a href=\"".hq("$selfurl?u=".uq($s)).$self->params_uq."\">".hq($s)."</a>";
	    }
	    my $checkbox;
	    if ($editable) {
		$checkbox = "<input type='checkbox' name='sp' value=\"".hq("$s $p")."\" /> ";
	    }
	    $resp->write("<tr>$c1<td>$checkbox$val</td></tr>\n");
	    $c1 = "";
	}
    }
    $resp->write("</table>\n");
    if ($editable) {
	$resp->write("<input type='submit' name='delete' value='Delete selected' />
<dl>
<dt>Predicate</dt>
  <dd><input type='text' name='p' width='60' /></dd>
<dt>Object</dt>
  <dd><input type='text' name='o' width='60' />
   <input type='submit' name='add' value='Add' />
  </dd>
  <dd><textarea name='ol' cols='60' rows='4'></textarea>
   <input type='submit' name='add.lit' value='Add (literal)' />
  </dd>
</dl>
</form>\n");
    }
    # Other queries: Goto URI and search
    $self->trailing_forms($query);
    $resp->write($query->end_html);
}

#-----------------------------------------------------------------------
#	Render trailing html forms
#-----------------------------------------------------------------------
sub trailing_forms {
    my($self, $query) = @_;
    my $resp = $query->response;
    my $hidden = $self->params_hidden;
    $resp->write("<p /><form method=\"GET\">
$hidden
<input type=\"text\" name=\"u\" value=\"\" />
<input type=\"submit\" value=\"Go to URI\" />
</form>\n");
    $resp->write("<p /><form method=\"GET\">
$hidden
<input type=\"text\" name=\"search\" value=\"\" />
<input type=\"submit\" value=\"Subject and literal search\" />
</form>\n");
}

#-----------------------------------------------------------------------
#	Search for a string in subjects and literals
#-----------------------------------------------------------------------
sub search {
    my($self, $query) = @_;
    my $search = $query->param('search');
    if ($search !~ /\S/) {
	die "Empty search string\n";
    }
    my $resp = $query->response;
    my $selfurl = $query->url(-path_info=>1);

    $resp->write($query->start_html(-title=>"Search for [$search]"));

    $resp->write(" <table border='1'>\n");
    if ($self->{'backparam'}) {
	my $backurl = $self->{'backparam'};
        my $hback = "<a href=\"".hq($backurl)."\">".hq($backurl)."</a>";
        $resp->write(" <tr><th colspan='2'>Model context: $hback</th></tr>\n");
    }

    # Look through everything
    my %matching_subjects;
    my %subjects_with_matching_literals;
    foreach my $stmt ($self->{'db'}->query(undef,undef,undef)) {
	my($subj,$pred,$obj) = @$stmt;
	if (index($subj,$search) >= 0) {
	    $matching_subjects{$subj} = 1;
	}
	if ($obj =~ /^=.*\Q$search\E/) {
	    $subjects_with_matching_literals{$subj} = 1;
	}
    }
    if (%matching_subjects == 0) {
	$resp->write("<tr><th colspan='2'>No subjects matching search for [$search].</th><tr>");
    } else {
	$resp->write("<tr><th>Subjects matching search<br /> for [$search]:</th><td>
<ul>\n");
	foreach my $subj (sort keys %matching_subjects) {
	    my $val = "<li><a href=\"" .
	      hq("$selfurl?u=".uq($subj).$self->params_uq) .
		"\">".hq($subj) .  "</a></li>\n";
	    $resp->write($val);
	}
	$resp->write("</ul>\n</td></tr>\n");
    }
    if (%subjects_with_matching_literals == 0) {
	$resp->write("<tr><th colspan='2'>No literals matching search for [$search].</th><tr>");
    } else {
	$resp->write("<tr><th>Subjects with literals<br />matching search for [$search]:</th><td>
<ul>\n");
	foreach my $subj (sort keys %subjects_with_matching_literals) {
	    my $val = "<li><a href=\"" .
	      hq("$selfurl?u=".uq($subj).$self->params_uq) .
		"\">".hq($subj) . "</a></li>\n";
	    $resp->write($val);
	}
	$resp->write("</ul>\n</td></tr>\n");
    }
    $resp->write("</table>\n");
    $self->trailing_forms($query);
    $resp->write($query->end_html);
}

#-----------------------------------------------------------------------
#	Delete some statements
#-----------------------------------------------------------------------
sub delete {
    my($self, $query,$uri) = @_;
    my $db = $self->{'db'};

    # Delete forward relations
    my @po = $query->param('po');
    foreach (@po) {
	my($p,$o) = split / /;
	if ($o =~ /^=/) {
	    # Literal as MD5 checksum
	    foreach my $lit ($db->sp2o($uri,$p)) {
		if ($lit =~ /^=/ &&
		    md5_hex(substr($lit,1)) eq substr($o,1)) {
		    $db->remove($uri,$p,$lit);
		}
	    }
	} else {
	    # Resource URI
	    $db->remove($uri,$p,$o);
	}
    }
    # Delete backward relations
    my @po = $query->param('sp');
    foreach (@po) {
	my($s,$p) = split / /;
	$db->remove($s,$p,$uri);
    }
    $db->commit;
}

#-----------------------------------------------------------------------
#	Add a statement
#-----------------------------------------------------------------------
sub add {
    my($self, $query,$uri,$literal) = @_;
    my $db = $self->{'db'};
    my($p,$o);
    $p = $query->param('p');
    die "Malformed URI\n"
	if $p =~ /\s/;
    if ($literal) {
	$o = "=" . $query->param('ol');
    } else {
	$o = $query->param('o');
	die "Malformed URI\n"
	    if $o =~ /\s/;
    }
    $db->add($uri,$p,$o);
    $db->commit;
}

#-----------------------------------------------------------------------
#	Return query context parameters as hidden form parameters
#-----------------------------------------------------------------------
sub params_hidden {
    my($self) = @_;
    my $hidden;
    if ($self->{'dbparam'}) {
	$hidden .= "<input type='hidden' name='db' value=\"".hq($self->{'dbparam'})."\" />\n";
    }
    if ($self->{'backparam'}) {
	$hidden .= "<input type='hidden' name='back' value=\"".hq($self->{'backparam'})."\" />\n";
    }
    return $hidden;
}

#-----------------------------------------------------------------------
#	Return query context parameters uri quoted
#-----------------------------------------------------------------------
sub params_uq {
    my($self) = @_;
    my $paramuq;
    if ($self->{'dbparam'}) {
        $paramuq = "&db=".uq($self->{'dbparam'}); # reproduce db param on new queries
    }
    if ($self->{'backparam'}) {
        $paramuq .= "&back=".uq($self->{'backparam'}); # reproduce back param on new queries
    }
    return $paramuq;
}

#-----------------------------------------------------------------------
#	Utility routines
#-----------------------------------------------------------------------
sub user_in_list {
    my($user, $list) = @_;
    if (! defined $list) {
	# No list, so nobody in it
	return undef;
    } elsif ($list eq '*') {
	# Means everybody
	return 1;
    } elsif (ref($list) eq 'ARRAY') {
	# Simple list of usernames
	return (grep {$user eq $_} @$list) != 0;
    } elsif (ref($list) eq 'CODE') {
	# Some algorithm
	return $list->($user);
    } else {
	die "user_in_list: Unrecognised list: $list";
    }
}

1;
